;; Copyright (C) 2004 Johannes Grdem <johs@copyleft.no>
;; Redistribution and use in source and binary forms, with or without
;; modification, is permitted.

;; THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS
;; OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
;; WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
;; ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE
;; LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
;; CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
;; OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
;; BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
;; LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
;; (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
;; USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
;; DAMAGE.

(declare
 (export pg:connect pg:reset pg:close pg:query-fold-left pg:query-for-each
         pg:query-tuples pg:sql-null-object? pg:connection? pg:escape-string
	 pg:sql-null-object
	 pg:named-tuples)
 (uses lolevel srfi-1 srfi-13 srfi-18 extras srfi-4))

(use easyffi)

(foreign-declare "#include <libpq-fe.h>")
(foreign-parse
"enum ConnStatusType
{
 CONNECTION_OK, CONNECTION_BAD, CONNECTION_STARTED,
 CONNECTION_MADE, CONNECTION_AWAITING_RESPONSE,
 CONNECTION_AUTH_OK, CONNECTION_SETENV, CONNECTION_SSL_STARTUP,
 CONNECTION_NEEDED
};

enum ExecStatusType
{
 PGRES_EMPTY_QUERY = 0, PGRES_COMMAND_OK, PGRES_TUPLES_OK,
 PGRES_COPY_OUT, PGRES_COPY_IN, PGRES_BAD_RESPONSE,
 PGRES_NONFATAL_ERROR, PGRES_FATAL_ERROR
};

enum PostgresPollingStatusType
{
 PGRES_POLLING_FAILED = 0, PGRES_POLLING_READING, PGRES_POLLING_WRITING,
 PGRES_POLLING_OK, PGRES_POLLING_ACTIVE
};

typedef unsigned int Oid;

PGconn   *PQconnectStart (const char *conninfo);
enum PostgresPollingStatusType PQconnectPoll (PGconn *conn);
int       PQsocket (const PGconn *conn);
int       PQresetStart (PGconn *conn);
enum PostgresPollingStatusType PQresetPoll (PGconn *conn);
void      PQfinish (PGconn *conn);
enum      ConnStatusType PQstatus (const PGconn *conn);
int       PQsendQuery (PGconn *conn, const char *query);
int PQsendQueryParams(PGconn *conn, const char *command,
                      int nParams, const Oid *paramTypes,
                      const char **paramValues, const int *paramLengths,
                      const int *paramFormats, int resultFormat);
int       PQisBusy (PGconn *conn);
int       PQconsumeInput (PGconn *conn);
PGresult *PQgetResult (PGconn *conn);
enum ExecStatusType PQresultStatus (const PGresult *res);
char     *PQresultErrorMessage (const PGresult *res);
char     *PQerrorMessage(const PGconn *conn);
void      PQclear (PGresult *res);
int       PQntuples (const PGresult *res);
int       PQnfields (const PGresult *res);
char     *PQcmdTuples (PGresult *res);
Oid       PQftype (const PGresult *res, int column_number);
char     *PQfname(const PGresult *res, int field_index);
void     *PQgetvalue (const PGresult *res, int row_number, int column_number);
int       PQgetisnull (const PGresult *res, int row_number, int column_number);
int       PQgetlength (const PGresult *res, int row_number,
                             int column_number);
char     *PQresultErrorField(const PGresult *res, int fieldcode);
"
)

(define-macro (define-foreign-int name)
  `(define-foreign-variable ,name int ,(conc "(int) " name)))

(define-foreign-int PG_DIAG_SEVERITY)
(define-foreign-int PG_DIAG_SQLSTATE)
(define-foreign-int PG_DIAG_MESSAGE_PRIMARY)
(define-foreign-int PG_DIAG_MESSAGE_DETAIL)
(define-foreign-int PG_DIAG_MESSAGE_HINT)
(define-foreign-int PG_DIAG_STATEMENT_POSITION)
(define-foreign-int PG_DIAG_CONTEXT)
(define-foreign-int PG_DIAG_SOURCE_FILE)
(define-foreign-int PG_DIAG_SOURCE_LINE)
(define-foreign-int PG_DIAG_SOURCE_FUNCTION)

(define (pg:error loc message . args)
  (signal (make-pg-condition loc message args: args)))

(define (make-pg-condition loc message #!key (args '()) severity error-class error-code message-detail
			   message-hint statement-position context source-file source-line
			   source-function)
  (make-composite-condition
    (make-property-condition 'exn 'location loc 'message message 'arguments args)
    (make-property-condition 'postgresql 'severity severity 'error-class error-class
			     'error-code error-code 'message-detail message-detail
			     'message-hint message-hint 'statement-position statement-position
			     'context context 'source-file source-file 'source-line source-line
			     ;; Might break not-terribly-old versions of postgresql
			     ;;'internal-position internal-position 'internal-query internal-query
 			     'source-function source-function)))

(define-record pg:connection
  ptr)

(define-record sql-null-object)
(define pg:sql-null-object (make-sql-null-object))

(define (pg:sql-null-object? x)
  (eq? x pg:sql-null-object))

(define-syntax push!
  (syntax-rules ()
    ((_ value place)
     (set! place (cons value place)))))

;; Block the current thread until there is input available on fd, if
;; input? is true, or block until it is possible to write to fd, if
;; input? is false.  (Thanks, Felix.)
(define (block-thread! fd input?)
  (let ([t (current-thread)])
    (##sys#thread-block-for-i/o! t fd input?)
    (##sys#call-with-current-continuation
     (lambda (return)
       (##sys#setslot t 1 (lambda () (return (void))))
       (##sys#schedule)))))

(define (pg:poll conn-ptr poll-function)
  (let ([conn-fd (PQsocket conn-ptr)])
    (let loop ([result (poll-function conn-ptr)])
      (cond [(= result PGRES_POLLING_OK)
             conn-ptr]
            [(= result PGRES_POLLING_FAILED)
	     (let ((error-message (PQerrorMessage conn-ptr)))
	       (PQfinish conn-ptr)
	       (pg:error 'pg:connect (conc "Polling Postgres database failed. "
					   error-message)))]
            [(member result (list PGRES_POLLING_WRITING PGRES_POLLING_READING))
             (block-thread! conn-fd (= PGRES_POLLING_READING result))
             (loop (poll-function conn-ptr))]))))

(define pg:connect
  (let ([types-mapped? #f])
    (lambda (connection-spec)
      ;; This function will yield when connections aren't ready as long as
      ;; it doesn't have to do (blocking) DNS-lookup to connect to the
      ;; database:
      (define (pg-connect-nonblocking connection-spec)
        (let ([conn-ptr (PQconnectStart connection-spec)])
          (cond
           [(not conn-ptr)
            (pg:error 'pg:connect
                      "Unable to allocate a Postgres connection structure."
		      connection-spec)]
           [(= CONNECTION_BAD (PQstatus conn-ptr))
	    (let ((error-message (PQerrorMessage conn-ptr)))
	     (PQfinish conn-ptr)
	     (pg:error 'pg:connect (conc "Connection to Postgres database failed: " error-message)
		       connection-spec))]
           [else
            (pg:poll conn-ptr PQconnectPoll)])))
      
      (define (connspec-escape s)
        (string-translate* (->string s) '(("'" . "\\'"))))
      
      (let ([conn (make-pg:connection
                   (pg-connect-nonblocking
                    (string-intersperse
                     (map (lambda (subspec)
                            (sprintf "~A = '~A'" (car subspec)
                                    (connspec-escape (cdr subspec))))
                          connection-spec)
                     " ")))])

        (set-finalizer! conn
                        (lambda (conn)
                          (pg:close conn)))
        
        ;; Retrieve type-information from PostgreSQL metadata for use by the
        ;; various value-parsers.
        (unless types-mapped?
          (fixup-types conn)
          (set! types-mapped? #t))
        conn))))

(define (pg:reset connection)
  (let ([conn-ptr (pg:connection-ptr connection)])
    (case (PQresetStart conn-ptr)
      ((1)
       (pg:poll conn-ptr PQresetPoll)
       (void))
      ((0)
       (let ((error-message (PQerrorMessage conn-ptr)))
	(pg:error 'pg:reset (conc "Reset of connection failed " error-message)
		  connection))))))

(define (pg:close connection)
  (let ([conn-ptr (pg:connection-ptr connection)])
    (when conn-ptr
      (PQfinish conn-ptr)
      (pg:connection-ptr-set! connection #f)))
  (void))

;; XXX: Should probably provide escaping function somewhere.

;; Buffer all available input, yielding if nothing is available:
(define (buffer-available-input! conn-ptr query)
  (let ([query-reply (if (pair? query)
                         (pq-send-query-params conn-ptr query)
                         (PQsendQuery conn-ptr query))])

    (if (zero? query-reply)
        (pg:error 'buffer-available-input!
                  (sprintf "PQsendQuery: ~A"
			   (string-trim-right (PQerrorMessage conn-ptr)))
		  query conn-ptr))
  
    (let ([conn-fd (PQsocket conn-ptr)])
      (let loop ()
        (case (PQconsumeInput conn-ptr)
          ((1)
           (when (= (PQisBusy conn-ptr) 1)
             (block-thread! conn-fd #t)
             (loop)))
          ((0)
           (pg:error 'buffer-available-input!
                     (sprintf "PQconsumeInput: ~A"
                             (string-trim-right (PQerrorMessage conn-ptr)))
		     conn-ptr))))
      (void))))

;; Collect the result pointers from the last query:
(define (collect-results conn-ptr)
  (define (cleanup-results! conn-ptr)
    (let loop ()
      (let ([result (PQgetResult conn-ptr)])
        (when result
          (PQclear result)
          (loop))))
    (void))

  (let loop ([results (list)])
    (let ([result (PQgetResult conn-ptr)])
      (if result
          (cond
           [(member (PQresultStatus result) (list PGRES_BAD_RESPONSE
                                                  PGRES_FATAL_ERROR))
            (let* ([msg (string-trim-right (PQresultErrorMessage result))]
		   [get-error-field (lambda (diag) (PQresultErrorField result diag))]
		   [sqlstate (get-error-field PG_DIAG_SQLSTATE)]
		   [maybe-severity (get-error-field PG_DIAG_SEVERITY)]
		   [maybe-statement-position (get-error-field PG_DIAG_STATEMENT_POSITION)]
		   [condition
		    (make-pg-condition
		     'collect-results
		     (sprintf "PQgetResult: ~A" msg)
		     args:               (list conn-ptr)
		     severity:           (and maybe-severity
					      (string->symbol (string-downcase maybe-severity)))
		     error-class:         (and sqlstate (sqlstate->error-class sqlstate))
		     error-code:         (and sqlstate (sqlstate->error-code sqlstate))
		     message-detail:     (get-error-field PG_DIAG_MESSAGE_DETAIL)
		     message-hint:       (get-error-field PG_DIAG_MESSAGE_HINT)
		     statement-position:  (and maybe-statement-position
					       (string->number maybe-statement-position))
		     context:            (get-error-field PG_DIAG_CONTEXT)
		     source-file:        (get-error-field PG_DIAG_SOURCE_FILE)
		     source-line:        (get-error-field PG_DIAG_SOURCE_LINE)
		     source-function:    (get-error-field PG_DIAG_SOURCE_FUNCTION))])
              (PQclear result)
              (cleanup-results! conn-ptr)
	      (signal condition))]
           [else
            (loop (cons result results))])
          (reverse! results)))))

(define (make-value-parsers result nfields)
  (do ([col 0 (+ col 1)]
       [parsers (make-vector nfields)])
      ([= col nfields] parsers)
    (vector-set! parsers col
                 (hash-table-ref type-hash (PQftype result col)
                                 (lambda () text-parser)))))

(define pg:named-tuples (make-parameter #f))

(define (pg:query-fold-left query connection fold-function . seeds)
  (let ([conn-ptr (pg:connection-ptr connection)])

    (buffer-available-input! conn-ptr query)

    (let ([results (collect-results conn-ptr)]
          [seed-count (length seeds)])

      (map (lambda (result)
             (set-finalizer! result
                             (lambda (r)
                               (PQclear r))))
           results)
      
      (let process-result ([results results]
                           [seeds seeds])

        (if (null? results)
            (apply values seeds)
            (let* ([result (car results)]
                   [result-status (PQresultStatus result)])
              
              (cond
               ;; No tuples returned.  Instead of tuple, provide the
               ;; number of tuples affected by the query:
               [(= result-status PGRES_COMMAND_OK)
                (let ([ntuples (string->number (PQcmdTuples result))])
                  (receive (proceed? . new-seeds)
                      (apply fold-function ntuples seeds)
                    (cond [(= (length new-seeds) seed-count)
                           (if proceed?
                               (process-result (cdr results) new-seeds)
                               (apply values new-seeds))]
                          [else
                           (error "Wrong seed count"
                                  `(expected ,seed-count)
                                  `(got ,(length new-seeds)))])))]

               ;; 0 or more tuples were returned.
               [(= result-status PGRES_TUPLES_OK)
                (let* ([nfields (PQnfields result)]
                       [ntuples (PQntuples result)]
                       [value-parsers (make-value-parsers
                                       result nfields)])
              
                  (define (get-value row column)
                    (if (zero? (PQgetisnull result row column))
                        (let ([value (PQgetvalue result row column)]
                              [value-length
                               (PQgetlength result row column)])
                          ((vector-ref value-parsers column)
                           value value-length))
                        pg:sql-null-object ))
              
                  (let process-row ([row 0]
                                    [seeds seeds])

                    (if (= row ntuples)
                        (process-result (cdr results) seeds)
			 (let ([tuple (make-vector nfields)]
			       [get-value* (if (pg:named-tuples)
					       (lambda (row col)
						  (cons (string->symbol (PQfname result col))
							(get-value row col)))
					       get-value)])
			    
                          (do ([col 0 (+ col 1)])
                              ([= col nfields])
				(vector-set! tuple col (get-value* row col)))
                          
                          (receive (proceed? . new-seeds)
                              (apply fold-function tuple seeds)
                            (cond
                             [(= (length new-seeds) seed-count)
                              (if proceed?
                                  (process-row (+ row 1) new-seeds)
                                  (apply values new-seeds))]
                             [else
                              (error "Wrong seed count"
                                     `(expected ,seed-count)
                                     `(got ,(length new-seeds)))]))))))]

               [(= result-status PGRES_EMPTY_QUERY)
                (pg:error 'pg-query-fold-left "Empty query")]

               [else
                (pg:error 'pg-query-fold-left
                          "Unsupported result type:" result-status)])))))))

(define (pg:query-for-each proc query connection)
  (receive ()
      (pg:query-fold-left query connection
                          (lambda (tuple)
                            (proc tuple)
                            #t))
    (void)))

(define (pg:query-tuples query connection)
  (reverse!
   (pg:query-fold-left query connection
                       (lambda (tuple tuples)
                         (values #t (cons tuple tuples)))
                       '())))

(define (text-parser pointer len)
  (do ([i 0 (+ i 1)]
       [result-string (make-string len)]
       [value pointer (pointer-offset value 1)])
      ([= i len] result-string)
    (string-set! result-string i
                 (integer->char (pointer-u8-ref value)))))

(define (char-parser pointer len)
  (integer->char (pointer-u8-ref pointer)))

(define (bool-parser pointer len)
  (case (char-parser pointer len)
    ((#\t) #t)
    ((#\f) #f)))

(define (abstime-parser pointer len)
  (text-parser pointer len))

(define (reltime-parser pointer len)
  (text-parser pointer len))

(define (parse-format-string s)
  (do ([i 0 (+ i 1)]
       [ranges (list)]
       [cur-range (list)]
       [len (string-length s)])
      ([= i len]
       (when (not (null? cur-range))
         (push! (cons (- i (length cur-range)) i)
                ranges))
       (reverse! ranges))
    (let ([char (string-ref s i)])
      (cond ([and (or (null? cur-range)
                      (char=? char (car cur-range)))
                  (char-alphabetic? char)]
             (push! char cur-range))
            ([and (not (null? cur-range))
                  (not (char=? char (car cur-range)))]
             (push! (cons (- i (length cur-range)) i)
                    ranges)
             (set! cur-range
                   (if (char-alphabetic? char)
                       (list char)
                       (list))))))))

(define-syntax define-time-parser
  (syntax-rules ()
    ((_ name format-string)
     (define name
       (let ([format-ranges (parse-format-string format-string)])
         (lambda (pointer length)
           (let ([date-string (text-parser pointer length)])
             (apply
              vector
              (map (lambda (range)
                     (if (> (cdr range) (string-length date-string))
                         0
                         (string->number
                          (substring date-string (car range) (cdr range)))))
                   format-ranges)))))))))

(define-time-parser date-parser "YYYY-MM-DD")
(define-time-parser timestamp-parser "YYYY-MM-DD hh:mm:ss.ssssss")
(define-time-parser timestamp/tz-parser "YYYY-MM-DD hh:mm:ss.sssssszzz")
(define-time-parser time-parser "hh:mm:ss.ssssss")

(define (interval-parser pointer len)
  (text-parser pointer len))

(define (numeric-parser pointer len)
  (let* ([num-string (text-parser pointer len)]
         [num (string->number num-string)])
    (if num
        num
        (pg:error 'numeric-parser "Unable to parse the number" num-string))))

(define type-map
  `(("text" . ,text-parser)
    ("bytea" . ,text-parser)
    ("char" . ,char-parser)
    ("bpchar" . ,text-parser)
    ("bool" . ,bool-parser)
    ("int8" . ,numeric-parser)
    ("int4" . ,numeric-parser)
    ("int2" . ,numeric-parser)
    ("float4" . ,numeric-parser)
    ("float8" . ,numeric-parser)
    ("abstime" . ,abstime-parser)
    ("reltime" . ,reltime-parser)
    ("date" . ,date-parser)
    ("time" . ,time-parser)
    ("timestamp" . ,timestamp-parser)
    ("timestamptz" . ,timestamp/tz-parser)
    ("interval" . ,interval-parser)
    ("numeric" . ,numeric-parser)
    ("oid" . ,numeric-parser)))

(define type-hash (make-hash-table))

;; Retrieve type-oids from PostgreSQL:
(define (fixup-types connection)
  (pg:query-for-each
   (lambda (parameters)
     (let* ([oid (string->number (vector-ref parameters 0))]
            [typname (vector-ref parameters 1)]
            [procedure (assoc typname type-map)])
       (when procedure
         (hash-table-set! type-hash oid (cdr procedure)))))
   "SELECT oid, typname FROM pg_type"
   connection))


;; Escape strings
(define %pg:escape-string
  (foreign-lambda* c-string* ((c-string from) (int fromlen))
		   "char *to = malloc(sizeof(char)*fromlen*2+1);" ;;
                   "PQescapeString(to, from, fromlen);"
		   "return(to);"
		   ))

(define %pg:escape-string-conn
  (foreign-lambda* c-string* ((pointer conn) (c-string from) (int fromlen))
		   "int err = 0; char *to = malloc(sizeof(char)*fromlen*2+1);" ;;
                   "PQescapeStringConn(conn, to, from, fromlen, &err);"
		   "if (err == 1) {"
		   "  free(to);"
		   "  return(NULL);"
		   "}"
		   "return(to);"
		   ))

(define pg:escape-string
  (case-lambda
    ((str) (%pg:escape-string str (string-length str)))
    ((conn str) (or (%pg:escape-string-conn conn str (string-length str))
                    (pg:error 'pg:connect (conc "String escaping failed. " (PQerrorMessage conn)))))))


(define sqlstate->error-class
  (let ((dict '(("00" . successful-completion)
		("01" . warning)
		("02" . no-data)
		("03" . sql-statement-not-yet-complete)
		("08" . connection-exception)
		("09" . triggered-action-exception)
		("0A" . feature-not-supported)
		("0B" . invalid-transaction-initiation)
		("0F" . locator-exception)
		("0L" . invalid-grantor)
		("0P" . invalid-role-specification)
		("21" . cardinality-violation)
		("22" . data-exception)
		("23" . integrity-constraint-violation)
		("24" . invalid-cursor-state)
		("25" . invalid-transaction-state)
		("26" . invalid-sql-statement-name)
		("27" . triggered-data-change-violation)
		("28" . invalid-authorization-specification)
		("2B" . dependent-privilege-descriptors-still-exist)
		("2D" . invalid-transaction-termination)
		("2F" . sql-routine-exception)
		("34" . invalid-cursor-name)
		("38" . external-routine-exception)
		("39" . external-routine-invocation-exception)
		("3D" . invalid-catalog-name)
		("3F" . invalid-schema-name)
		("40" . transaction-rollback)
		("42" . syntax-error-or-access-rule-violation)
		("44" . with-check-option-violation)
		("53" . insufficient-resources)
		("54" . program-limit-exceeded)
		("55" . object-not-in-prerequisite-state)
		("57" . operator-intervention)
		("58" . system-error)
		("f0" . configuration-file-error)
		("XX" . internal-error))))
    (lambda (s)
      (cdr (assoc (string-take s 2) dict)))))



(define sqlstate->error-code
  (let ((dict '(("00000" . successful-completion)
		("01000" . warning)
		("0100C" . warning-dynamic-result-sets-returned)
		("01008" . warning-implicit-zero-bit-padding)
		("01003" . warning-null-value-eliminated-in-set-function)
		("01004" . warning-string-data-right-truncation)
		("02000" . no-data)
		("02001" . no-additional-dynamic-result-sets-returned)
		("03000" . sql-statement-not-yet-complete)
		("08000" . connection-exception)
		("08003" . connection-does-not-exist)
		("08006" . connection-failure)
		("08001" . sqlclient-unable-to-establish-sqlconnection)
		("08004" . sqlserver-rejected-establishment-of-sqlconnection)
		("08007" . transaction-resolution-unknown)
		("08P01" . protocol-violation)
		("09000" . triggered-action-exception)
		("0A000" . feature-not-supported)
		("0B000" . invalid-transaction-initiation)
		("0F000" . locator-exception)
		("0F001" . invalid-specification)
		("0L000" . invalid-grantor)
		("0LP01" . invalid-grant-operation)
		("0P000" . invalid-role-specification)
		("21000" . cardinality-violation)
		("22000" . data-exception)
		("2202E" . array-element-error)
		("22021" . character-not-in-repertoire)
		("22008" . datetime-field-overflow)
		("22012" . division-by-zero)
		("22005" . error-in-assignment)
		("2200B" . escape-character-conflict)
		("22022" . indicator-overflow)
		("22015" . interval-field-overflow)
		("22018" . invalid-character-value-for-cast)
		("22007" . invalid-datetime-format)
		("22019" . invalid-escape-character)
		("2200D" . invalid-escape-octet)
		("22025" . invalid-escape-sequence)
		("22010" . invalid-indicator-parameter-value)
		("22020" . invalid-limit-value)
		("22023" . invalid-parameter-value)
		("2201B" . invalid-regular-expression)
		("22009" . invalid-time-zone-displacement-value)
		("2200C" . invalid-use-of-escape-character)
		("2200G" . most-specific-type-mismatch)
		("22004" . null-value-not-allowed)
		("22002" . null-value-no-indicator-parameter)
		("22003" . numeric-value-out-of-range)
		("22026" . string-data-length-mismatch)
		("22001" . string-data-right-truncation)
		("22011" . substring-error)
		("22027" . trim-error)
		("22024" . unterminated-c-string)
		("2200F" . zero-length-character-string)
		("22P01" . floating-point-exception)
		("22P02" . invalid-text-representation)
		("22P03" . invalid-binary-representation)
		("22P04" . bad-copy-file-format)
		("22P05" . untranslatable-character)
		("23000" . integrity-constraint-violation)
		("23001" . restrict-violation)
		("23502" . not-null-violation)
		("23503" . foreign-key-violation)
		("23505" . unique-violation)
		("23514" . check-violation)
		("24000" . invalid-cursor-state)
		("25000" . invalid-transaction-state)
		("25001" . active-sql-transaction)
		("25002" . branch-transaction-already-active)
		("25008" . held-cursor-requires-same-isolation-level)
		("25003" . inappropriate-access-mode-for-branch-transaction)
		("25004" . inappropriate-isolation-level-for-branch-transaction)
		("25005" . no-active-sql-transaction-for-branch-transaction)
		("25006" . read-only-sql-transaction)
		("25007" . schema-and-data-statement-mixing-not-supported)
		("25P01" . no-active-sql-transaction)
		("25P02" . in-failed-sql-transaction)
		("26000" . invalid-sql-statement-name)
		("27000" . triggered-data-change-violation)
		("28000" . invalid-authorization-specification)
		("2B000" . dependent-privilege-descriptors-still-exist)
		("2BP01" . dependent-objects-still-exist)
		("2D000" . invalid-transaction-termination)
		("2F000" . sql-routine-exception)
		("2F005" . function-executed-no-return-statement)
		("2F002" . modifying-sql-data-not-permitted)
		("2F003" . prohibited-sql-statement-attempted)
		("2F004" . reading-sql-data-not-permitted)
		("34000" . invalid-cursor-name)
		("38000" . external-routine-exception)
		("38001" . containing-sql-not-permitted)
		("38002" . modifying-sql-data-not-permitted)
		("38003" . prohibited-sql-statement-attempted)
		("38004" . reading-sql-data-not-permitted)
		("39000" . external-routine-invocation-exception)
		("39001" . invalid-sqlstate-returned)
		("39004" . null-value-not-allowed)
		("39P01" . trigger-protocol-violated)
		("39P02" . srf-protocol-violated)
		("3D000" . invalid-catalog-name)
		("3F000" . invalid-schema-name)
		("40000" . transaction-rollback)
		("40002" . integrity-constraint-violation)
		("40001" . serialization-failure)
		("40003" . statement-completion-unknown)
		("40P01" . deadlock-detected)
		("42000" . syntax-error-or-access-rule-violation)
		("42601" . syntax-error)
		("42501" . insufficient-privilege)
		("42846" . cannot-coerce)
		("42803" . grouping-error)
		("42830" . invalid-foreign-key)
		("42602" . invalid-name)
		("42622" . name-too-long)
		("42939" . reserved-name)
		("42804" . datatype-mismatch)
		("42P18" . indeterminate-datatype)
		("42809" . wrong-object-type)
		("42703" . undefined-column)
		("42883" . undefined-function)
		("42P01" . undefined-table)
		("42P02" . undefined-parameter)
		("42704" . undefined-object)
		("42701" . duplicate-column)
		("42P03" . duplicate-cursor)
		("42P04" . duplicate-database)
		("42723" . duplicate-function)
		("42P05" . duplicate-pstatement)
		("42P06" . duplicate-schema)
		("42P07" . duplicate-table)
		("42712" . duplicate-alias)
		("42710" . duplicate-object)
		("42702" . ambiguous-column)
		("42725" . ambiguous-function)
		("42P08" . ambiguous-parameter)
		("42P09" . ambiguous-alias)
		("42P10" . invalid-column-reference)
		("42611" . invalid-column-definition)
		("42P11" . invalid-cursor-definition)
		("42P12" . invalid-database-definition)
		("42P13" . invalid-function-definition)
		("42P14" . invalid-pstatement-definition)
		("42P15" . invalid-schema-definition)
		("42P16" . invalid-table-definition)
		("42P17" . invalid-object-definition)
		("44000" . with-check-option-violation)
		("53000" . insufficient-resources)
		("53100" . disk-full)
		("53200" . out-of-memory)
		("53300" . too-many-connections)
		("54000" . program-limit-exceeded)
		("54001" . statement-too-complex)
		("54011" . too-many-columns)
		("54023" . too-many-arguments)
		("55000" . object-not-in-prerequisite-state)
		("55006" . object-in-use)
		("55P02" . cant-change-runtime-param)
		("57000" . operator-intervention)
		("57014" . query-canceled)
		("57P01" . admin-shutdown)
		("57P02" . crash-shutdown)
		("57P03" . cannot-connect-now)
		("58030" . io-error)
		("58P01" . undefined-file)
		("58P02" . duplicate-file)
		("F0000" . config-file-error)
		("F0001" . lock-file-exists)
		("XX000" . internal-error)
		("XX001" . data-corrupted)
		("XX002" . index-corrupted))))
    	 (lambda (s) (cdr (assoc s dict)))))



;;; query with params stuff

(define (pq-send-query-params conn query-args #!key (format 'text))
  ;; first attempt: all args will be text. I'll look at binary later.
  (let* ((query (car query-args))
         (args (map ->string (cdr query-args)))
         (pcount (length args))
         (pvalues (map ->string args))
         (pvalues (make-ppchar args))
         (plengths (list->s32vector (map string-length args)))
         (pformats (list->s32vector (map (constantly 0) args))) ; 1 for binary, 0 for text.
         (ptypes (list->u32vector (map (constantly 0) args)))
         (result (PQsendQueryParams conn query
                                    pcount ptypes pvalues plengths pformats
                                    (if (eq? format 'text) 0 1))))
    (free-ppchar pvalues)
    result))

(define make-ppchar (foreign-lambda (pointer (pointer char)) "make_ppchar" scheme-object))
(define free-ppchar (foreign-lambda void "free_ppchar" (pointer (pointer char))))
;;(define test-ppchar (foreign-lambda void "test_ppchar" (pointer (pointer char))))
